@@ -1,6 +1,7 @@
 #!/usr/bin/env python
 # encoding: utf-8
 
+import sys
 from Naked.settings import debug as DEBUG_FLAG
 
 #------------------------------------------------------------------------------
@@ -8,13 +9,15 @@ from Naked.settings import debug as DEBUG_FLAG
 #   A generic Python object
 #   Assigns object attributes by key name in the dictionary argument to the constructor
 #   The methods are inherited by other mutable Naked object extension types
+#   Attribute accessors: hasattr, getattr, setattr, delattr
 #------------------------------------------------------------------------------
-class NakedObject:
-    # initialize with an attributes dictionary {attribute_name, attribute_value}
-    def __init__(self, attributes={}):
+class NakedObject(object):
+    # initialize with an attributes dictionary {attribute_name: attribute_value}
+    def __init__(self, attributes={}, naked_type='NakedObject'):
         if len(attributes) > 0:
             for key in attributes:
                 setattr(self, key, attributes[key])
+        setattr(self, '_naked_type_', naked_type) # maintain an attribute to keep track of the extension type
 
     #------------------------------------------------------------------------------
     # [ _getAttributeDict method ] (dictionary)
@@ -24,37 +27,61 @@ class NakedObject:
         return self.__dict__
 
     #------------------------------------------------------------------------------
-    # [ getAttribute method ] (attribute dependent type)
-    #  returns the respective attribute for the `attribute` name on the NakedObject instance
+    # [ _equal_type method ] (boolean)
+    #  returns boolean for type of instance == type of test parameter instance
     #------------------------------------------------------------------------------
-    def getAttribute(self, attribute):
-        if hasattr(self, attribute):
-            return getattr(self, attribute)
+    def _equal_type(self, other_obj):
+        return type(self) == type(other_obj)
+
+    #------------------------------------------------------------------------------
+    # [ _equal_attributes metod ] (method)
+    #  returns boolean for instance.__dict__ == test parameter .__dict__ (attribute comparison)
+    #------------------------------------------------------------------------------
+    def _equal_attributes(self, other_obj):
+        return self.__dict__ == other_obj.__dict__
+
+    #------------------------------------------------------------------------------
+    # == overload
+    #------------------------------------------------------------------------------
+    def __eq__(self, other_obj):
+        return self.equals(other_obj)
+
+    #------------------------------------------------------------------------------
+    # != overload
+    #------------------------------------------------------------------------------
+    def __ne__(self, other_obj):
+        result = self.equals(other_obj)
+        if result:
+            return False # reverse result of the equals method
+        else:
+            return True
+
+    #------------------------------------------------------------------------------
+    # [ equals method ] (boolean)
+    #   equality testing based on type and attributes
+    #   **NEED TO OVERRIDE IN CLASSES THAT INHERIT
+    #------------------------------------------------------------------------------
+    def equals(self, other_obj):
+        return self._equal_type(other_obj) and self._equal_attributes(other_obj)
+
+    #------------------------------------------------------------------------------
+    # [ type method ] (string)
+    #  returns the Naked type extension string that is set in the constructor for each object type
+    #------------------------------------------------------------------------------
+    def type(self):
+        if hasattr(self, '_naked_type_'):
+            return self._naked_type_
         else:
             return None
 
-    #------------------------------------------------------------------------------
-    # [ setAttribute method ] (no return value)
-    #  sets a NakedObject attribute `value` for the `attribute` name
-    #------------------------------------------------------------------------------
-    def setAttribute(self, attribute, value):
-        setattr(self, attribute, value)
-
-    #------------------------------------------------------------------------------
-    # [ hasAttribute method ] (boolean)
-    #  returns truth test for presence of an `attribute` name on the NakedObject
-    #------------------------------------------------------------------------------
-    def hasAttribute(self, attribute):
-        return hasattr(self, attribute)
-
 #------------------------------------------------------------------------------
 # [[ XDict class ]]
-#   An inherited extension to the dictionary object that permits attachment of attributes
+#   An inherited extension to the dictionary type
 #------------------------------------------------------------------------------
 class XDict(dict, NakedObject):
-    def __init__(self, dict_obj, attributes={}):
+    def __init__(self, dict_obj, attributes={}, naked_type='XDict'):
         dict.__init__(self, dict_obj)
-        NakedObject.__init__(self, attributes)
+        NakedObject.__init__(self, attributes, naked_type)
 
     #------------------------------------------------------------------------------
     # XDict Operator Overloads
@@ -66,25 +93,66 @@ class XDict(dict, NakedObject):
     #   returns the updated XDict object
     #------------------------------------------------------------------------------
     def __add__(self, other_dict):
-        self.update(other_dict)
-        return self
+        try:
+            self.update(other_dict)
+            if hasattr(other_dict, '_naked_type_') and (getattr(other_dict, '_naked_type_') == 'XDict'):
+                attr_dict = other_dict._getAttributeDict() # get the attributes from the parameter XDict and add to new XDict
+                if len(attr_dict) > 0:
+                    for key in attr_dict:
+                        setattr(self, key, attr_dict[key])
+            return self
+        except Exception as e:
+            if DEBUG_FLAG:
+                sys.stderr.write("Naked Framework Error: unable to combine XDict with parameter provided (Naked.toolshed.types.py)")
+            raise e
 
     #------------------------------------------------------------------------------
     #  +- overload
     #  overwrites existing keys with another_dict (right sided argument) keys if they are the same keys
     #  returns the updated XDict object
     #------------------------------------------------------------------------------
-    def __iadd__(self, another_dict):
-        self.update(another_dict)
-        return self
+    def __iadd__(self, other_dict):
+        try:
+            self.update(other_dict)
+            if hasattr(other_dict, '_naked_type_') and (getattr(other_dict, '_naked_type_') == 'XDict'):
+                attr_dict = other_dict._getAttributeDict() # get the attributes from the parameter XDict and add to new XDict
+                if len(attr_dict) > 0:
+                    for key in attr_dict:
+                        setattr(self, key, attr_dict[key])
+            return self
+        except Exception as e:
+            if DEBUG_FLAG:
+                sys.stderr.write("Naked Framework Error: unable to combine XDict with parameter provided (Naked.toolshed.types.py)")
+            raise e
 
     #------------------------------------------------------------------------------
-    # << overload extends XDict with another dictionary
-    #  overwrites existing keys with another_dict (right sided argument) keys if they are the same
+    # == overload
     #------------------------------------------------------------------------------
-    def __lshift__(self, another_dict):
-        self.update(another_dict)
-        return self
+    def __eq__(self, other_obj):
+        return self.equals(other_obj)
+
+    #------------------------------------------------------------------------------
+    # != overload
+    #------------------------------------------------------------------------------
+    def __ne__(self, other_obj):
+        result = self.equals(other_obj)
+        if result:
+            return False # reverse result of the equals method
+        else:
+            return True
+
+    #------------------------------------------------------------------------------
+    # [ equals method ] (boolean)
+    #   tests for equality of the XDict (type, attributes, dictionary equality)
+    #------------------------------------------------------------------------------
+    def equals(self, other_obj):
+        if self._equal_type(other_obj) and self._equal_attributes(other_obj):
+            if dict(self) == dict(other_obj):
+                return True
+            else:
+                return False
+        else:
+            return False
 
     #------------------------------------------------------------------------------
     # XDict Value Methods
@@ -93,6 +161,7 @@ class XDict(dict, NakedObject):
     # [ conditional_map_to_vals method ] (XDict)
     #  returns the original XDict with values that meet True condition in `conditional_function`
     #  modified as per the `mapped_function` with single value argument call
+    #  Test: test_xdict_conditional_map
     #------------------------------------------------------------------------------
     def conditional_map_to_vals(self, conditional_function, mapped_function):
         for key, value in self.xitems():
@@ -103,6 +172,7 @@ class XDict(dict, NakedObject):
     #------------------------------------------------------------------------------
     # [ map_to_vals method ] (XDict)
     #  returns the original XDict with all values modified as per the `mapped_function`
+    #  Test: test_xdict_map_to_vals
     #------------------------------------------------------------------------------
     def map_to_vals(self, mapped_function):
         # return XDict( zip(self, map(mapped_function, self.values())), self._getAttributeDict() ) - slower in Py2
@@ -113,12 +183,14 @@ class XDict(dict, NakedObject):
     #------------------------------------------------------------------------------
     # [ val_xlist method ] (XList)
     #  return an XList of the values in the XDict
+    #  Test: test_xdict_val_xlist
     #------------------------------------------------------------------------------
     def val_xlist(self):
         return XList(self.values(), self._getAttributeDict())
 
     #------------------------------------------------------------------------------
     # [ max_val method ] (tuple of maximum value and associated key)
+    #  Test: test_xdict_max_val, test_xdict_max_val_strings (strings are alphabetic if not numerals)
     #------------------------------------------------------------------------------
     def max_val(self):
         return max(zip(self.values(), self.keys()))
@@ -139,27 +211,30 @@ class XDict(dict, NakedObject):
     #------------------------------------------------------------------------------
     # [ val_count method ] (integer)
     #  returns an integer value for the total count of `value_name` in the dictionary values
-    #  Case sensitive test for strings
+    #  Case sensitive test if comparing strings
+    #  Tests: test_xdict_val_count_string, test_xdict_val_count_integer
     #------------------------------------------------------------------------------
     def val_count(self, value_name):
         count = 0
-        for test_string in self.values():
-            if value_name == test_string:
+        for test_value in self.values():
+            if value_name == test_value:
                 count += 1
         return count
 
     #------------------------------------------------------------------------------
     # [ value_count_ci method ] (integer)
     #  returns an integer value for the total count of case insensitive `value_name`
-    #  strings/char in the dictionary values.  This is a string only search method
+    #  strings/char in the dictionary values.  Can include non-string types (ignores them)
+    #  Test: test_xdict_val_count_ci
     #------------------------------------------------------------------------------
     def val_count_ci(self, value_name):
-        value_name = value_name.lower()
-        lower_case_vals = [ x.lower() for x in self.values() ]
         count = 0
-        for test_string in lower_case_vals:
-            if value_name in test_string:
-                count += 1
+        for test_value in self.values():
+            try:
+                if value_name.lower() in test_value.lower():
+                    count += 1
+            except AttributeError: # the test_value was not a string, catch exception and continue count attempt
+                continue
         return count
 
 
@@ -169,6 +244,7 @@ class XDict(dict, NakedObject):
     #------------------------------------------------------------------------------
     # [ difference method ] (difference set of keys)
     #  definition: keys that are included in self, but not in `another_dict`
+    #  Tests: test_xdict_key_difference, test_xdict_key_difference_when_none_present
     #------------------------------------------------------------------------------
     def difference(self, another_dict):
         return set(self.keys()) - set(another_dict.keys())
@@ -176,6 +252,7 @@ class XDict(dict, NakedObject):
     #------------------------------------------------------------------------------
     # [ intersection method ] (intersection set of keys)
     #   definition: keys that are included in both self and `another_dict`
+    #   Tests: test_xdict_key_intersection, test_xdict_key_intersection_when_none_present
     #------------------------------------------------------------------------------
     def intersection(self, another_dict):
         return set(self.keys()) & set(another_dict.keys())
@@ -183,6 +260,7 @@ class XDict(dict, NakedObject):
     #------------------------------------------------------------------------------
     # [ key_xlist method ] (XList)
     #  returns an XList of the keys in the XDict
+    #  Test: test_xdict_key_xlist
     #------------------------------------------------------------------------------
     def key_xlist(self):
         return XList(self.keys(), self._getAttributeDict())
@@ -190,19 +268,23 @@ class XDict(dict, NakedObject):
     #------------------------------------------------------------------------------
     # [ random method ] (dictionary)
     #  return new Python dictionary with single, random key:value pair
+    #  Test: test_xdict_key_random
     #------------------------------------------------------------------------------
     def random(self):
         import random
-        random_key = random.choice(self.keys())
-        return {random_key: self[random_key]}
+        from Naked.toolshed.python import py_major_version
+        random_key_list = random.sample(self.keys(), 1)
+        the_key = random_key_list[0]
+        return {the_key: self[the_key]}
 
     #------------------------------------------------------------------------------
     # [ random_sample method ] (dictionary)
     #  return new Python dictionary with `number_of_items` random key:value pairs
+    #  Test: test_xdict_key_random_sample
     #------------------------------------------------------------------------------
     def random_sample(self, number_of_items):
         import random
-        random_key_list = random.sample(self, number_of_items)
+        random_key_list = random.sample(self.keys(), number_of_items)
         new_dict = {}
         for item in random_key_list:
             new_dict[item] = self[item]
@@ -212,6 +294,7 @@ class XDict(dict, NakedObject):
     # [ xitems method ] (tuple)
     #   Generator method that returns tuples of every key, value in dictionary
     #   uses appropriate method from Python 2 and 3 interpreters
+    #   Test: test_xdict_xitems
     #------------------------------------------------------------------------------
     def xitems(self):
         from Naked.toolshed.python import py_major_version
@@ -225,9 +308,9 @@ class XDict(dict, NakedObject):
 #  An inherited extension to the list object that permits attachment of attributes
 #------------------------------------------------------------------------------
 class XList(list, NakedObject):
-    def __init__(self, list_obj, attributes={}):
+    def __init__(self, list_obj, attributes={}, naked_type='XList'):
         list.__init__(self, list_obj)
-        NakedObject.__init__(self, attributes)
+        NakedObject.__init__(self, attributes, naked_type)
 
     #------------------------------------------------------------------------------
     # XList Operator Overloads
@@ -500,8 +583,8 @@ class XList(list, NakedObject):
 #------------------------------------------------------------------------------
 from heapq import heappush, heappop
 class XPriorityQueue(NakedObject):
-    def __init__(self, initial_iterable=[], attributes={}):
-        NakedObject.__init__(self, attributes)
+    def __init__(self, initial_iterable=[], attributes={}, naked_type='XPriorityQueue'):
+        NakedObject.__init__(self, attributes, naked_type)
         self._queue = []
         self._index = 0
 
@@ -527,9 +610,9 @@ class XPriorityQueue(NakedObject):
 #------------------------------------------------------------------------------
 from collections import deque
 class XQueue(deque, NakedObject):
-    def __init__(self, initial_iterable=[], attributes={}, max_length=10):
+    def __init__(self, initial_iterable=[], attributes={}, max_length=10, naked_type='XQueue'):
         deque.__init__(self, initial_iterable, max_length)
-        NakedObject.__init__(self, attributes)
+        NakedObject.__init__(self, attributes, naked_type)
 
 
 #------------------------------------------------------------------------------
@@ -538,9 +621,9 @@ class XQueue(deque, NakedObject):
 #  Inherits from set and from NakedObject (see methods in NakedObject at top of this module
 #------------------------------------------------------------------------------
 class XSet(set, NakedObject):
-    def __init__(self, set_obj, attributes={}):
+    def __init__(self, set_obj, attributes={}, naked_type='XSet'):
         set.__init__(self, set_obj)
-        NakedObject.__init__(self, attributes)
+        NakedObject.__init__(self, attributes, naked_type)
 
     #   << operator is overloaded to extend the XSet with a second set
     def __lshift__(self, another_set):
@@ -566,29 +649,34 @@ class XSet(set, NakedObject):
 #  Immutable so there is no setter method, attributes must be set in the constructor
 #------------------------------------------------------------------------------
 class XFSet(frozenset):
-    def __new__(cls, the_set, attributes={}):
+    def __new__(cls, the_set, attributes={}, naked_type="XFSet"):
         set_obj = frozenset.__new__(cls, the_set)
         if len(attributes) > 0:
             for key in attributes:
                 setattr(set_obj, key, attributes[key])
+        setattr(set_obj, '_naked_type_', naked_type) # set the naked extension type as an attribute (NakedObject does this for mutable classes)
         return set_obj
 
     def _getAttributeDict(self):
         return self.__dict__
 
-    def getAttribute(self, attribute):
-        if hasattr(self, attribute):
-            return getattr(self, attribute)
-        else:
-            return None
-
     def xlist(self):
         attr_dict = self._getAttributeDict()
-        return XList(list(self), attr_dict)
+        return XList(list(self), attr_dict, naked_type="XList")
 
     def xset(self):
         attr_dict = self._getAttributeDict()
-        return XSet(self, attr_dict)
+        return XSet(self, attr_dict, naked_type="XSet")
+
+    #------------------------------------------------------------------------------
+    # [ type method ] (string)
+    #  returns the Naked type extension string that is set in the constructor for each object type
+    #------------------------------------------------------------------------------
+    def type(self):
+        if hasattr(self, '_naked_type_'):
+            return self._naked_type_
+        else:
+            return None
 
 
 #------------------------------------------------------------------------------
@@ -599,16 +687,28 @@ class XFSet(frozenset):
 #   Python 3: string (that permits unicode) by default, can normalize with XString().unicode() method
 #------------------------------------------------------------------------------
 class XString(str):
-    def __new__(cls, string_text, attributes={}):
+    def __new__(cls, string_text, attributes={}, naked_type='XString'):
         str_obj = str.__new__(cls, string_text)
         if len(attributes) > 0:
             for key in attributes:
                 setattr(str_obj, key, attributes[key])
+        setattr(str_obj, '_naked_type_', naked_type)
         return str_obj
 
-    def getAttribute(self, attribute):
-        if hasattr(self, attribute):
-            return getattr(self, attribute)
+    #------------------------------------------------------------------------------
+    # [ _getAttributeDict method ] (dictionary)
+    #  returns a dictionary of the XString instance attributes
+    #------------------------------------------------------------------------------
+    def _getAttributeDict(self):
+        return self.__dict__
+
+    #------------------------------------------------------------------------------
+    # [ type method ] (string)
+    #  returns the Naked type extension string that is set in the constructor for each object type
+    #------------------------------------------------------------------------------
+    def type(self):
+        if hasattr(self, '_naked_type_'):
+            return self._naked_type_
         else:
             return None
 
@@ -659,7 +759,7 @@ class XString(str):
         from fnmatch import fnmatchcase
         return fnmatchcase(self, wildcard)
 
-    # convert string to normalized UTF-8 in Python 2 and 3
+    # convert string to normalized UTF-8 in Python 2 and 3 (##TODO: convert to XUnicode with attributes?)
     def unicode(self):
         from sys import version_info
         from unicodedata import normalize
@@ -671,25 +771,27 @@ class XString(str):
 
 # this version works
 class XUnicode:
-    def __init__(self, string_text, attributes={}):
+    def __init__(self, string_text, attributes={}, naked_type='XUnicode'):
         import sys
         import unicodedata
         norm_text = unicodedata.normalize('NFKD', string_text)
 
         class XUnicode_2(unicode):
-            def __new__(cls, the_string_text, attributes={}):
+            def __new__(cls, the_string_text, attributes={}, naked_type='XUnicode2'):
                 str_obj = unicode.__new__(cls, the_string_text)
                 if len(attributes) > 0:
                     for key in attributes:
                         setattr(str_obj, key, attributes[key])
+                setattr(str_obj, '_naked_type_', naked_type) # set the type to XUnicode2 for Py 2 strings
                 return str_obj
 
         class XUnicode_3(str):
-            def __new__(cls, the_string_text, attributes={}):
+            def __new__(cls, the_string_text, attributes={}, naked_type='XUnicode3'):
                 str_obj = str.__new__(cls, the_string_text)
                 if len(attributes) > 0:
                     for key in attributes:
                         setattr(str_obj, key, attributes[key])
+                setattr(str_obj, '_naked_type_', naked_type) # set the type to XUnicode3 for Py 3 strings
                 return str_obj
 
 
@@ -712,6 +814,24 @@ class XUnicode:
 
     def __cmp__(self, other_string):
         return hash(self.naked_u_string) ==  hash(other_string)
+        # TODO: add check for same attributes
+
+    #------------------------------------------------------------------------------
+    # [ _getAttributeDict method ] (dictionary)
+    #  returns a dictionary of the NakedObject instance attributes
+    #------------------------------------------------------------------------------
+    def _getAttributeDict(self):
+        return self.__dict__
+
+    #------------------------------------------------------------------------------
+    # [ type method ] (string)
+    #  returns the Naked type extension string that is set in the constructor for each object type
+    #------------------------------------------------------------------------------
+    def type(self):
+        if hasattr(self, '_naked_type_'):
+            return self._naked_type_
+        else:
+            return None
 
 
 #------------------------------------------------------------------------------
@@ -719,13 +839,31 @@ class XUnicode:
 #
 #------------------------------------------------------------------------------
 class XTuple(tuple):
-    def __new__(cls, the_tuple, attributes={}):
+    def __new__(cls, the_tuple, attributes={}, naked_type='XTuple'):
         tup_obj = tuple.__new__(cls, the_tuple)
         if len(attributes) > 0:
             for key in attributes:
                 setattr(tup_obj, key, attributes[key])
+        setattr(tup_obj, '_naked_type_', naked_type)
         return tup_obj
 
+    #------------------------------------------------------------------------------
+    # [ _getAttributeDict method ] (dictionary)
+    #  returns a dictionary of the NakedObject instance attributes
+    #------------------------------------------------------------------------------
+    def _getAttributeDict(self):
+        return self.__dict__
+
+    #------------------------------------------------------------------------------
+    # [ type method ] (string)
+    #  returns the Naked type extension string that is set in the constructor for each object type
+    #------------------------------------------------------------------------------
+    def type(self):
+        if hasattr(self, '_naked_type_'):
+            return self._naked_type_
+        else:
+            return None
+
 
 if __name__ == '__main__':
     pass
